local debug = false
--[===[@debug@
debug = true
--@end-debug@]===]
local printdelay = {}
local function dprint(...)
	if (not debug) then return end
	local o
	for i=1, select('#',...) do
		o=(o and o.." " or "")..tostring((select(i,...)))
	end
	if (not printdelay[o]) or (printdelay[o] <= GetTime()) then
		print("[CBMT_SpellLogger Debug] "..o)
		printdelay[o] = GetTime()+5
	end
end

local is42 = ((select(4,GetBuildInfo()))==40200)
local function clogarg(tag)
	if tag == 'time' then
		return 1
	end
	if tag == 'subevent' then
		return 2
	end
	if tag == 'srcguid' then
		return 4
	end
	if tag == 'srcname' then
		return 5
	end
	if tag == 'dstguid' then
		if is42 then return 8 else return 7 end
	end
	if tag == 'dstname' then
		if is42 then return 9 else return 8 end
	end
	if tag == 'spellid' then
		if is42 then return 12 else return 10 end
	end
	if tag == 'spellname' then
		if is42 then return 13 else return 11 end
	end
	error('Invalid CL tag: '..tag)
end

local ignore_npcs = { -- because guid doesn't differentiate between guardian and "true" npc, fml
	[2523] = true, -- searing totem
	[5874] = true, -- strength of earth totem
	[5873] = true, -- stoneskin totem
	[2630] = true, -- earthbind totem
	[5913] = true, -- tremor totem
	[15430] = true, -- earth elemental totem
	[15352] = true, -- greater earth elemental
	[3579] = true, -- stoneclaw totem
	[6112] = true, -- windfury totem
	[5925] = true, -- grounding totem
	[15447] = true, -- wrath of air
	[3527] = true, -- healing stream totem
	[3573] = true, -- mana spring totem
	[5927] = true, -- elemental resistance totem
	[10467] = true, -- mana tide totem
	[5950] = true, -- flametongue totem
	[5929] = true, -- magma totem
	[15439] = true, -- fire elemental totem
	[15438] = true, -- greater fire elemental
	[46954] = true, -- shadowy apparition
	[19668] = true, -- shadowfiend
	[31216] = true, -- mirror image
	[27829] = true, -- gargoyle
	[46157] = true, -- hand of gul'dan
	[30230] = true, -- risen ally [REDUNDANT 4.1]
}

local scanTip = CreateFrame("GameTooltip", "CBMTSpellLoggerSpellScanTooltip", UIParent, "GameTooltipTemplate")
local scanLine
local function GetSpellTooltipText(spellID)
	scanTip:SetOwner(UIParent, "ANCHOR_NONE")
	scanTip:SetSpellByID(spellID)
	scanLine = CBMTSpellLoggerSpellScanTooltipTextLeft3
	local t = scanLine:GetText()
	if (not t) then return end
	if (t == SPELL_CAST_TIME_INSTANT_NO_MANA) or (t:find(SPELL_CAST_TIME_SEC:gsub("%%%.3g", "%d+"))) then
		scanLine = CBMTSpellLoggerSpellScanTooltipTextLeft4
	end
	local r=scanLine:GetText():match("[ ]*(.+)")
	scanTip:Hide()
	return r
end

local function guid2npcid(guid)
	return tonumber("0x"..(guid:sub(7,10))) or 0
end

local function timeformat(sec)
	local timeinfo = date("!*t", sec)
	return (timeinfo.hour > 0) and ("%02d:%02d:%02d"):format(timeinfo.hour,timeinfo.min,timeinfo.sec) or ("%02d:%02d"):format(timeinfo.min,timeinfo.sec)
end

local tempTitleTable = {}
local function generateLogTitle(l)
	local startTime, endTime = l.startTime or l[1].t, l.endTime or l[#l].t
	local dateStart, dateEnd, timeStart, timeEnd = date("%m/%d/%y", startTime), date("%m/%d/%y", endTime), date("%H:%M:%S", startTime), date("%H:%M:%S", endTime)
	local title = "["..dateStart..(dateStart ~= dateEnd and "-"..dateEnd or "").." "..timeStart.."-"..timeEnd.."] "
	table.wipe(tempTitleTable)
	for name in pairs(l.names) do
		tinsert(tempTitleTable, name)
	end
	title = title..table.concat(tempTitleTable,", ")
	return title
end

local main = CreateFrame("Frame","CBMT_SpellLoggerEvent")
local logs, indexlogs
local isLogging = false
main:SetScript("OnEvent", function(self, event, ...)
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		local exists = false
		for bossNum=1, 4 do
			if UnitExists("boss"..bossNum) then
				exists = true
				break
			end
		end
		if not exists then
			self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
			isLogging = false
			logs[#logs].endTime = select(clogarg('time'),...)
			dprint("boss1-4 doesn't exist, quit logging")
			return
		end
		if isLogging and not logs[#logs].startTime then
			dprint("Start time for combat logged: "..(select(clogarg('time'),...)))
			logs[#logs].startTime = select(clogarg('time'),...)
		end
		--dprint("boss1 exists, named: "..UnitName("boss1"))
		local srcGUID = select(clogarg('srcguid'), ...)
		if bit.band(srcGUID:sub(1,5),0x00F) == 0x003 and not ignore_npcs[guid2npcid(srcGUID)] then
			local type = select(clogarg('subevent'), ...)
			if type == "SPELL_CAST_START" or type == "SPELL_CAST_SUCCESS" then
				tinsert(logs[#logs], {
					t=select(clogarg('time'),...),
					src = select(clogarg('srcname'), ...),
					dst = select(clogarg('dstname'), ...),
					srcGUID = srcGUID,
					dstGUID = select(clogarg('dstguid'), ...),
					spellID = select(clogarg('spellid'), ...),
					spellName = select(clogarg('spellname'), ...),
				})
			end
			logs[#logs].startTimes = logs[#logs].startTimes or {}
			logs[#logs].startTimes[srcGUID] = logs[#logs].startTimes[srcGUID] or select(clogarg('time'),...)
		end
		local dstGUID = select(clogarg('dstguid'),...)
		if bit.band(dstGUID:sub(1,5),0x00F) == 0x003 and not ignore_npcs[guid2npcid(dstGUID)] then
			logs[#logs].startTimes = logs[#logs].startTimes or {}
			logs[#logs].startTimes[dstGUID] = logs[#logs].startTimes[dstGUID] or select(clogarg('time'),...)
		end
	elseif event == "ADDON_LOADED" then
		local n = select(1, ...)
		if n == "CBMT" or n == "CBMT_SpellLogger" then
			CBMT.opt.logs = CBMT.opt.logs or {}
			CBMT.opt.indexedfights = CBMT.opt.indexedfights or {}
			logs = CBMT.opt.logs
			indexlogs = CBMT.opt.indexedfights
			self:UnregisterEvent("ADDON_LOADED")
		end
	elseif event == "INSTANCE_ENCOUNTER_ENGAGE_UNIT" then
		local index = 1
		while UnitExists("boss"..index) do
			local name = UnitName("boss"..index)
			if not ((not name) or (name == UNKNOWN)) then
				if isLogging then
					logs[#logs].names[name] = true
					dprint("Boss engaged: "..name.."; Already logging, add to names list.")
				else
					tinsert(logs, {
						names={
							[name] = true,
						},
					})
					self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
					isLogging = true
					dprint("Boss engaged: "..name.."; Start logging.")
				end
			end
			index=index+1
		end
	end
end)
main:RegisterEvent("ADDON_LOADED")
main:RegisterEvent("INSTANCE_ENCOUNTER_ENGAGE_UNIT")


local display = CreateFrame("Frame", "CBMT_SpellLogger")
display:SetFrameStrata("DIALOG")
display:SetSize(600,400)
display:SetPoint("CENTER")
display:SetMovable(true)
display:Hide()
display.backdrop = display:CreateTexture()
display.backdrop:SetAllPoints()
display.backdrop:SetTexture(0.3,0.3,0.3)
display.backdrop:SetAlpha(0.5)
display.header = CreateFrame("Frame", nil, display)
display.header:SetPoint("TOPLEFT")
display.header:SetPoint("TOPRIGHT")
display.header:SetHeight(20)
display.header:SetScript("OnMouseDown", function() display:StartMoving() end)
display.header:SetScript("OnMouseUp", function() display:StopMovingOrSizing() end)
display.header.texture = display.header:CreateTexture()
display.header.texture:SetTexture(0.05,0.05,0.5)
display.header.texture:SetAllPoints()
display.header.text = display.header:CreateFontString()
display.header.text:SetTextColor(0.8,0.8,0.8)
display.header.text:SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE")
display.header.text:SetPoint("LEFT", 5, 0)
display.header.text:SetText("CBMT: SpellLogger")
display.header.close = CreateFrame("Button", nil, display, "UIPanelCloseButton")
display.header.close:SetPoint("TOPRIGHT", display.header, "TOPRIGHT", 2.5, 2.5)
display.header.close:SetSize(25,25)
display.header.close:SetFrameLevel(display.header:GetFrameLevel()+1)
display.tabs = CreateFrame("Frame", nil, display)
display.tabs:SetPoint("TOPLEFT", display.header, "BOTTOMLEFT")
display.tabs:SetPoint("TOPRIGHT", display.header, "BOTTOMRIGHT")
display.tabs:SetHeight(30)
display.overlay = CreateFrame("Frame", nil, display)
display.overlay:SetFrameLevel(display:GetFrameLevel()+2)
display.overlay.tasks = {}
display.overlay:EnableMouse(true)
display.overlay:SetPoint("TOPLEFT", display.header, "BOTTOMLEFT")
display.overlay:SetPoint("BOTTOMRIGHT")
display.overlay.display = CreateFrame("Frame", nil, display.overlay)
display.overlay.display:SetPoint("CENTER")
display.overlay.display:SetSize(500, 70)
display.overlay.display.header = display.overlay.display:CreateTexture()
display.overlay.display.header:SetPoint("TOPLEFT")
display.overlay.display.header:SetPoint("TOPRIGHT")
display.overlay.display.header:SetHeight(20)
display.overlay.display.header:SetTexture(0.05,0.05,0.5)
display.overlay.display.text = display.overlay.display:CreateFontString()
display.overlay.display.text:SetPoint("CENTER", 0, -10)
display.overlay.display.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
display.overlay.display.text:SetTextColor(1,1,1)
display.overlay.display.backdrop = display.overlay.display:CreateTexture()
display.overlay.display.backdrop:SetAllPoints()
display.overlay.display.backdrop:SetTexture(0.1,0.1,0.1)
display.overlay.display.backdrop:SetAlpha(0.5)

local function checkLog(log)
	if not log.isIndexed then
		if #log >= 1 then
			tinsert(display.overlay.tasks, {log,'all'})
		end
		display.overlay:Show()
	else
		if not log.isIndexed.spell then
			tinsert(display.overlay.tasks, {log,'spell'})
			display.overlay:Show()
		end
		if not log.isIndexed.spellindexing then
			tinsert(display.overlay.tasks, {log,'spellindexing'})
			display.overlay:Show()
		end
		if not log.isIndexed.npc then
			tinsert(display.overlay.tasks, {log,'npc'})
			display.overlay:Show()
		end
		if not log.isIndexed.npcindexing then
			tinsert(display.overlay.tasks, {log,'npcindexing'})
			display.overlay:Show()
		end
	end
end

display.overlay:SetScript("OnUpdate", function(self)
	if not self.tasks[1] then self:Hide() return end
	local log, type = unpack(self.tasks[1])
	log.title = log.title or generateLogTitle(log)
	if type == 'all' then
		self.display.text:SetText("Generating index list\n"..log.title)
		log.isIndexed = {}
		checkLog(log)
	end
	if type == 'spell' then
		self.display.text:SetText("Adding log to spells index\n"..log.title)
		indexlogs.spell = indexlogs.spell or {}
		for _, data in ipairs(log) do
			indexlogs.spell[data.spellID] = indexlogs.spell[data.spellID] or {}
			indexlogs.spell[data.spellID][log.title] = true
		end
		log.isIndexed.spell = true
	end
	if type == 'spellindexing' then
		self.display.text:SetText("Generating spells index for log\n"..log.title)
		log.indexes = log.indexes or {}
		log.indexes.spell = {}
		for index, data in ipairs(log) do
			log.indexes.spell[data.spellID] = log.indexes.spell[data.spellID] or {n=data.spellName}
			log.indexes.spell[data.spellID][data.srcGUID] = log.indexes.spell[data.spellID][data.srcGUID] or {n=data.src}
			if not tContains(log.indexes.spell[data.spellID][data.srcGUID], index) then
				tinsert(log.indexes.spell[data.spellID][data.srcGUID], index)
			end
		end
		log.isIndexed.spellindexing = true
	end
	if type == 'npc' then
		self.display.text:SetText("Adding log to NPC index\n"..log.title)
		indexlogs.npc = indexlogs.npc or {}
		for _, data in ipairs(log) do
			indexlogs.npc[data.src] = indexlogs.npc[data.src] or {}
			indexlogs.npc[data.src][log.title] = true
		end
		log.isIndexed.npc = true
	end
	if type == 'npcindexing' then
		self.display.text:SetText("Generating NPC index for log\n"..log.title)
		log.indexes = log.indexes or {}
		log.indexes.npc = {}
		for index, data in ipairs(log) do
			log.indexes.npc[data.src] = log.indexes.npc[data.src] or {}
			log.indexes.npc[data.src][data.srcGUID] = log.indexes.npc[data.src][data.srcGUID] or {}
			if not tContains(log.indexes.npc[data.src][data.srcGUID], index) then
				tinsert(log.indexes.npc[data.src][data.srcGUID], index)
			end
		end
		log.isIndexed.npcindexing = true
	end
	tremove(self.tasks, 1)
end)

display:SetScript("OnShow", function()
	for n, log in ipairs(logs) do
		checkLog(log)
	end
end)

local spellViewers = setmetatable({}, {__index=function(t, key)
	local spellviewer = CreateFrame("Frame", "CBMT_SpellViewer_"..key, display)
	spellviewer:SetFrameStrata("FULLSCREEN_DIALOG")
	spellviewer:SetSize(500, 350)
	spellviewer:SetPoint("CENTER")
	spellviewer:SetMovable(true)
	spellviewer:Hide()
	spellviewer.backdrop = spellviewer:CreateTexture()
	spellviewer.backdrop:SetAllPoints()
	spellviewer.backdrop:SetTexture(0.3,0.3,0.3)
	--spellviewer.backdrop:SetAlpha(0.5)
	spellviewer.header = CreateFrame("Frame", nil, spellviewer)
	spellviewer.header:SetPoint("TOPLEFT")
	spellviewer.header:SetPoint("TOPRIGHT")
	spellviewer.header:SetHeight(20)
	spellviewer.header:SetScript("OnMouseDown", function() spellviewer:StartMoving() end)
	spellviewer.header:SetScript("OnMouseUp", function() spellviewer:StopMovingOrSizing() end)
	spellviewer.header.texture = spellviewer.header:CreateTexture()
	spellviewer.header.texture:SetTexture(0.05,0.05,0.5)
	spellviewer.header.texture:SetAllPoints()
	spellviewer.header.text = spellviewer.header:CreateFontString()
	spellviewer.header.text:SetTextColor(0.8,0.8,0.8)
	spellviewer.header.text:SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE")
	spellviewer.header.text:SetPoint("LEFT", 5, 0)
	--spellviewer.header.text:SetText("View log:")
	spellviewer.header.close = CreateFrame("Button", nil, spellviewer, "UIPanelCloseButton")
	spellviewer.header.close:SetPoint("TOPRIGHT", spellviewer.header, "TOPRIGHT", 2.5, 2.5)
	spellviewer.header.close:SetSize(25,25)
	spellviewer.header.close:SetFrameLevel(spellviewer.header:GetFrameLevel()+1)
	spellviewer.tabBar = CreateFrame("Frame", nil, spellviewer)
	spellviewer.tabBar:SetPoint("TOPLEFT", spellviewer.header, "BOTTOMLEFT")
	spellviewer.tabBar:SetPoint("TOPRIGHT", spellviewer.header, "BOTTOMRIGHT")
	spellviewer.tabBar:SetHeight(25)
	spellviewer.tabBar.tabs = setmetatable({},{__index=function(tab,key)
		local f = CreateFrame("Button", nil, spellviewer.tabBar)
		f.label = f:CreateFontString()
		f.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
		f.label:SetTextColor(1,1,1)
		if key > 1 then
			f.label:SetPoint("TOPLEFT", tab[key-1].label, "TOPRIGHT", 5, 0)
			f.label:SetPoint("BOTTOMLEFT", tab[key-1].label, "BOTTOMRIGHT", 5, 0)
		else
			f.label:SetPoint("TOPLEFT", spellviewer.tabBar, "TOPLEFT", 5, -5)
			f.label:SetPoint("BOTTOMLEFT", spellviewer.tabBar, "BOTTOMLEFT", 5, 5)
		end
		f:SetPoint("TOPLEFT", f.label, "TOPLEFT", -5, 5)
		f:SetPoint("BOTTOMRIGHT", f.label, "BOTTOMRIGHT", 5, -5)
		f:RegisterForClicks("LeftButtonUp")
		f.highlight = f:CreateTexture()
		f.highlight:SetAllPoints()
		f.highlight:SetTexture(0.6,0.6,0.6)
		f.highlight:SetAlpha(0.3)
		f.highlight:Hide()
		f.selected = f:CreateTexture()
		f.selected:SetAllPoints()
		f.selected:SetTexture(0.8,0.8,0.8)
		f.selected:SetAlpha(0.6)
		f.selected:Hide()
		f:SetScript("OnEnter", function(self) if spellviewer.tabBar.selected ~= self then self.highlight:Show() end end)
		f:SetScript("OnLeave", function(self) self.highlight:Hide() end)
		f:SetScript("OnClick", function(self) if spellviewer.tabBar.selected then spellviewer.tabBar.selected.selected:Hide() spellviewer.tabBar.selected.frame:Hide() end self.clickFunc() self.frame:Show() self.selected:Show() spellviewer.tabBar.selected = self end)
		f.frame = CreateFrame("Frame", nil, spellviewer)
		f.frame:SetPoint("TOPLEFT", spellviewer.tabBar, "BOTTOMLEFT")
		f.frame:SetPoint("BOTTOMRIGHT")
		f.frame.backing = f.frame:CreateTexture()
		f.frame.backing:SetAllPoints()
		f.frame.backing:SetTexture(0.8,0.8,0.8)
		f.frame.backing:SetAlpha(0.6)
		f.frame.labels = setmetatable({},{__index=function(par,ind)
			local label = CreateFrame("Frame", nil, f.frame)
			label:SetHeight(20)
			label.line1 = label:CreateTexture()
			label.line1:SetTexture(0.8,0.8,0.8)
			label.line1:SetPoint("TOPLEFT", label, "LEFT", 0, -2)
			label.line1:SetPoint("TOPRIGHT", label, "RIGHT", 0, -2)
			label.line1:SetHeight(2)
			label.line1:SetAlpha(0.5)
			label.line2 = label:CreateTexture()
			label.line2:SetTexture(0.8,0.8,0.8)
			label.line2:SetPoint("BOTTOMLEFT", 0, 9)
			label.line2:SetPoint("BOTTOMRIGHT", 0, 9)
			label.line2:SetHeight(2)
			label.line2:SetAlpha(0.5)
			label.text = label:CreateFontString()
			label.text:SetFont("Fonts\\FRIZQT__.TTF", 14, "OUTLINE")
			label.text:SetTextColor(1,1,1)
			label.text:SetPoint("LEFT", 10, 0)
			function label:reset()
				label.text:SetText("")
				label:Hide()
				label:ClearAllPoints()
				label.used = false
			end
			function label:set(text)
				label.text:SetText(tostring(text))
				label:Show()
			end
			par[ind] = label
			return label
		end})
		f.frame.rows = setmetatable({},{__index=function(par,ind)
			local row = CreateFrame("Frame", nil, f.frame)
			row:SetHeight(15)
			row.label = row:CreateFontString()
			row.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
			row.label:SetTextColor(1,0.82,0)
			row.label:SetPoint("LEFT", 5, 0)
			row.label:SetJustifyH("LEFT")
			row.value = row:CreateFontString()
			row.value:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
			row.value:SetTextColor(1,1,1)
			row.value:SetPoint("RIGHT", -5, 0)
			row.value:SetJustifyH("RIGHT")
			function row:reset()
				row.label:SetText("")
				row.value:SetText("")
				row:Hide()
				row:ClearAllPoints()
				row.used = false
			end
			function row:set(label, value)
				row.label:SetText(tostring(label))
				row.value:SetText(tostring(value))
				row:Show()
			end
			par[ind] = row
			return row
		end})
		f.frame.children = {}
		function f.frame:reset()
			for _, c in ipairs(self.children) do
				c:reset()
			end
			table.wipe(self.children)
		end
		function f.frame:add(type)
			if not self[type.."s"] then return end
			local c=1
			while true do
				local x = self[type.."s"][c]
				if not tContains(self.children,x) then
					if #self.children >= 1 then
						x:SetPoint("TOPLEFT", self.children[#self.children], "BOTTOMLEFT")
						x:SetPoint("TOPRIGHT", self.children[#self.children], "BOTTOMRIGHT")
					else
						x:SetPoint("TOPLEFT")
						x:SetPoint("TOPRIGHT")
					end
					tinsert(self.children,x)
					x.used = true
					return x
				end
				c=c+1
			end
		end
		f.frame:Hide()
		function f:reset()
			if spellviewer.tabBar.selected == self then spellviewer.tabBar.selected = nil self.frame:Hide() self.frame:reset() self.selected:Hide() end
		end
		tab[key] = f
		return f
	end})
	function spellviewer:reset()
		for _, tab in ipairs(self.tabBar.tabs) do
			tab:reset()
		end
		self.header.text:SetText(nil)
	end
	t[key] = spellviewer
	return spellviewer
end})

local tempCalcTable= {}
local openSpell = function(spellID)
	local viewer
	for _, v in ipairs(spellViewers) do
		if not v:IsShown() then
			viewer = v
			break
		end
	end
	if not viewer then
		viewer = spellViewers[#spellViewers+1]
	end
	viewer:reset()
	viewer.header.text:SetText("Viewing spell: "..(GetSpellInfo(spellID)))
	viewer:Show()
	do
		local summary = viewer.tabBar.tabs[1]
		summary.frame:reset()
		summary.label:SetText("Timing information")
		summary.clickFunc = function()
		end
		table.wipe(tempCalcTable)
		local toIndex = indexlogs.spell[spellID]
		for _, log in ipairs(logs) do
			if toIndex[log.title] then
				local logindex = log.indexes.spell[spellID]
				local firstEntry = log.startTime or log[1].t
				if logindex then
					for guid, data in pairs(logindex) do
						if guid ~= "n" then
							local nid = guid2npcid(guid)
							tempCalcTable[nid] = tempCalcTable[nid] or {n=data.n, timediff={}, firstcast=math.huge}
							for i=2, #data do
								tinsert(tempCalcTable[nid].timediff, log[data[i]].t-log[data[i-1]].t)
							end
							if log.startTimes and log.startTimes[guid] then
								tempCalcTable[nid].firstcast = math.min(tempCalcTable[nid].firstcast, log[data[1]].t-log.startTimes[guid])
							end
						end
					end
				end
			end
		end
		for _, data in pairs(tempCalcTable) do
			summary.frame:add("label"):set(data.n)
			if data.firstcast < math.huge then
				summary.frame:add("row"):set("Time until first cast:", timeformat(data.firstcast))
			end
			if #data.timediff > 0 then
				summary.frame:add("row"):set("Minimum time between casts:", timeformat(math.min(unpack(data.timediff))))
				local sum=0
				for _, t in ipairs(data.timediff) do
					sum=sum+t
				end
				summary.frame:add("row"):set("Average time between casts:", timeformat(sum/#data.timediff))
				summary.frame:add("row"):set("Maximum time between casts:", timeformat(math.max(unpack(data.timediff))))
			end
		end
	end
	--[[for _, f in ipairs(viewer.leftScroll.entries) do
		f:Hide()
	end
	local i=1
	for spellid, data in pairs(log.indexes.spell) do
		print(spellid)
		viewer.leftScroll.entries[i].label:SetText("|T"..GetSpellTexture(spellid)..":16|t "..data.n)
		viewer.leftScroll.entries[i].data = data
		viewer.leftScroll.entries[i]:Show()
		i=i+1
	end
	viewer.leftScroll.child:SetHeight(20*(i-1))
	viewer.leftScroll.entries[1].label:SetText("Summary")
	viewer.leftScroll:UpdateScrollChildRect()
	viewer.leftScroll.entries[1]:Click()--]]
end

local openLog = function(log)
	print(log)
end

display.scrolls = {}
display.scrolls[3] = CreateFrame("ScrollFrame", nil, display)
display.scrolls[3]:SetPoint("TOPRIGHT", display.header, "BOTTOMRIGHT", -10, -10)
display.scrolls[3]:SetPoint("BOTTOMRIGHT", display, -10, 10)
display.scrolls[3].tex = display.scrolls[3]:CreateTexture(nil, "BACKGROUND")
display.scrolls[3].tex:SetAllPoints()
display.scrolls[3].tex:SetTexture(0.1,0.1,0.1)
display.scrolls[3].tex:SetAlpha(0.5)
display.scrolls[3]:SetWidth((display:GetWidth()-40)/3)
display.scrolls[3].child = CreateFrame("Frame", nil, display)
display.scrolls[3].child:EnableMouseWheel(true)
display.scrolls[3].child:SetScript("OnMouseWheel", function(self, delta)
	if delta == 1 then
		display.scrolls[3]:SetVerticalScroll(math.max(display.scrolls[3]:GetVerticalScroll()-(IsShiftKeyDown() and 40 or 10),0))
	else
		display.scrolls[3]:SetVerticalScroll(math.min(display.scrolls[3]:GetVerticalScroll()+(IsShiftKeyDown() and 40 or 10), display.scrolls[3]:GetVerticalScrollRange()))
	end
end)
display.scrolls[3].child:SetWidth(display.scrolls[3]:GetWidth())
display.scrolls[3]:SetScrollChild(display.scrolls[3].child)
display.scrolls[3].entries = setmetatable({}, {__index=function(tab, key)
	local f = CreateFrame("Button", nil, display.scrolls[3].child)
	if key > 1 then
		f:SetPoint("TOPLEFT", tab[key-1], "BOTTOMLEFT")
		f:SetPoint("TOPRIGHT", tab[key-1], "BOTTOMRIGHT")
	else
		f:SetPoint("TOPLEFT")
		f:SetPoint("TOPRIGHT")
	end
	f:SetHeight(30)
	f.highlight = f:CreateTexture()
	f.highlight:SetAllPoints()
	f.highlight:SetTexture(0.6,0.6,0.6)
	f.highlight:SetAlpha(0.3)
	f.highlight:Hide()
	f.selected = f:CreateTexture()
	f.selected:SetAllPoints()
	f.selected:SetTexture(0.8,0.8,0.8)
	f.selected:SetAlpha(0.6)
	f.selected:Hide()
	f:RegisterForClicks("LeftButtonUp")
	f:SetScript("OnEnter", function(self) if display.scrolls[3].selected ~= self then self.highlight:Show() end end)
	f:SetScript("OnLeave", function(self) self.highlight:Hide() end)
	f:SetScript("OnClick", function(self)
		if self.type == "log" then
			openLog(self.data)
		end
		if self.type == "spell" then
			openSpell(self.data)
		end
		if display.scrolls[3].selected then
			display.scrolls[3].selected.selected:Hide()
		end
		display.scrolls[3].selected = self
		self.selected:Show()
	end)
	f.icon = f:CreateTexture()
	f.icon:SetPoint("TOPLEFT", 5, -5)
	f.icon:SetPoint("BOTTOMLEFT", 5, 5)
	f.icon:SetWidth(20)
	f.textf = CreateFrame("Frame", nil, f)
	f.textf:SetPoint("TOPLEFT", 5, -5)
	f.textf:SetPoint("BOTTOMRIGHT", -5, 5)
	f.label = f.textf:CreateFontString()
	f.label:SetPoint("TOPLEFT")
	f.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	f.label:SetTextColor(1,1,1)
	f.desc = f.textf:CreateFontString()
	f.desc:SetPoint("BOTTOMLEFT")
	f.desc:SetPoint("BOTTOMRIGHT")
	f.desc:SetWordWrap(false)
	f.desc:SetJustifyH("LEFT")
	f.desc:SetFont("Fonts\\FRIZQT__.TTF", 8, "OUTLINE")
	f.desc:SetTextColor(1,1,1)
	function f:SetOpen(data, type)
		if type == "spell" then
			self.label:SetText((GetSpellInfo(data)).." ["..data.."]")
			local tt = GetSpellTooltipText(data)
			self.desc:SetText(tt and tt:gsub("\n"," "):match("[ ]*(.+)") or "")
			self.icon:SetTexture(GetSpellTexture(data))
			self.icon:Show()
			self.textf:ClearAllPoints()
			self.textf:SetPoint("TOPLEFT", self.icon, "TOPRIGHT", 5, 0)
			self.textf:SetPoint("BOTTOMRIGHT", -5, 5)
		elseif type == "log" then
			table.wipe(tempTitleTable)
			for name in pairs(data.names) do
				tinsert(tempTitleTable, name)
			end
			self.label:SetText(table.concat(tempTitleTable, "/"))
			self.desc:SetText("["..date("%m/%d/%y", data[1].t).."] Duration: "..math.ceil(data[#data].t-data[1].t).." sec")
			self.textf:ClearAllPoints()
			self.textf:SetPoint("TOPLEFT", 5, -5)
			self.textf:SetPoint("BOTTOMRIGHT", -5, 5)
			self.icon:Hide()
		end
		self.type = type
		self.data = data
	end
	--[[function f:SetLog(log)
		table.wipe(tempTitleTable)
		for name in pairs(log.names) do
			tinsert(tempTitleTable, name)
		end
		self.label:SetText(table.concat(tempTitleTable, "/"))
		self.desc:SetText("["..date("%m/%d/%y", log[1].t).."] Duration: "..math.ceil(log[#log].t-log[1].t).." sec")
		self.log = log
	end
	function f:SetData(data)
		self.data = data
	end--]]
	tab[key] = f
	return f
end})
display.scrolls[3]:SetScript("OnSizeChanged", function(self, width)
	self.child:SetWidth(width)
	self:UpdateScrollChildRect()
end)

display.scrolls[2] = CreateFrame("ScrollFrame", nil, display)
display.scrolls[2]:SetPoint("TOP", display.header, "BOTTOM", 0, -10)
display.scrolls[2]:SetPoint("BOTTOM", display, 0, 10)
display.scrolls[2].tex = display.scrolls[2]:CreateTexture(nil, "BACKGROUND")
display.scrolls[2].tex:SetAllPoints()
display.scrolls[2].tex:SetTexture(0.1,0.1,0.1)
display.scrolls[2].tex:SetAlpha(0.5)
display.scrolls[2]:SetWidth((display:GetWidth()-40)/3)
display.scrolls[2].child = CreateFrame("Frame", nil, display)
display.scrolls[2].child:EnableMouseWheel(true)
display.scrolls[2].child:SetScript("OnMouseWheel", function(self, delta)
	if delta == 1 then
		display.scrolls[2]:SetVerticalScroll(math.max(display.scrolls[2]:GetVerticalScroll()-10,0))
	else
		display.scrolls[2]:SetVerticalScroll(math.min(display.scrolls[2]:GetVerticalScroll()+10, display.scrolls[2]:GetVerticalScrollRange()))
	end
end)
display.scrolls[2].child:SetWidth(display.scrolls[2]:GetWidth())
display.scrolls[2]:SetScrollChild(display.scrolls[2].child)
display.scrolls[2].entries = setmetatable({}, {__index=function(tab, key)
	local f = CreateFrame("Button", nil, display.scrolls[2].child)
	if key > 1 then
		f:SetPoint("TOPLEFT", tab[key-1], "BOTTOMLEFT")
		f:SetPoint("TOPRIGHT", tab[key-1], "BOTTOMRIGHT")
	else
		f:SetPoint("TOPLEFT")
		f:SetPoint("TOPRIGHT")
	end
	f:SetHeight(20)
	f.highlight = f:CreateTexture()
	f.highlight:SetAllPoints()
	f.highlight:SetTexture(0.6,0.6,0.6)
	f.highlight:SetAlpha(0.3)
	f.highlight:Hide()
	f.selected = f:CreateTexture()
	f.selected:SetAllPoints()
	f.selected:SetTexture(0.8,0.8,0.8)
	f.selected:SetAlpha(0.6)
	f.selected:Hide()
	f:RegisterForClicks("LeftButtonUp")
	f:SetScript("OnEnter", function(self) if display.scrolls[2].selected ~= self then self.highlight:Show() end end)
	f:SetScript("OnLeave", function(self) self.highlight:Hide() end)
	f:SetScript("OnClick", function(self)
		for _, btn in ipairs(display.scrolls[3].entries) do btn:Hide() end
		if display.scrolls[3].selected then display.scrolls[3].selected.selected:Hide() end
		local i = 1
		for _, log in pairs(logs) do
			if log.title and self.list[log.title] then
				local btn = display.scrolls[3].entries[i]
				btn:SetOpen(log, "log")
				btn:Show()
				i=i+1
			end
		end
		if display.scrolls[2].selected then display.scrolls[2].selected.selected:Hide() end
		display.scrolls[2].selected = self
		display.scrolls[3].child:SetHeight((i-1)*30)
		display.scrolls[3]:UpdateScrollChildRect()
		self.selected:Show()
	end)
	f.label = f:CreateFontString()
	f.label:SetPoint("TOPLEFT", 5, -5)
	f.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	f.label:SetTextColor(1,1,1)
	function f:SetList(label,list)
		f.label:SetText(label)
		f.list = list
	end
	tab[key] = f
	return f
end})
display.scrolls[2]:SetScript("OnSizeChanged", function(self, width)
	self.child:SetWidth(width)
	self:UpdateScrollChildRect()
end)


display.scrolls[1] = CreateFrame("ScrollFrame", nil, display)
display.scrolls[1]:SetPoint("TOPLEFT", display.header, "BOTTOMLEFT", 10, -10)
display.scrolls[1]:SetPoint("BOTTOMLEFT", display, 10, 10)
display.scrolls[1].tex = display.scrolls[1]:CreateTexture(nil, "BACKGROUND")
display.scrolls[1].tex:SetAllPoints()
display.scrolls[1].tex:SetTexture(0.1,0.1,0.1)
display.scrolls[1].tex:SetAlpha(0.5)
display.scrolls[1]:SetWidth((display:GetWidth()-40)/3)
display.scrolls[1].child = CreateFrame("Frame", nil, display)
display.scrolls[1].child:EnableMouseWheel(true)
display.scrolls[1].child:SetScript("OnMouseWheel", function(self, delta)
	if delta == 1 then
		display.scrolls[1]:SetVerticalScroll(math.max(display.scrolls[1]:GetVerticalScroll()-10,0))
	else
		display.scrolls[1]:SetVerticalScroll(math.min(display.scrolls[1]:GetVerticalScroll()+10, display.scrolls[1]:GetVerticalScrollRange()))
	end
end)
display.scrolls[1].child:SetWidth(display.scrolls[1]:GetWidth())
display.scrolls[1]:SetScrollChild(display.scrolls[1].child)
display.scrolls[1].entries = setmetatable({}, {__index=function(tab, key)
	local f = CreateFrame("Button", nil, display.scrolls[1].child)
	if key > 1 then
		f:SetPoint("TOPLEFT", tab[key-1], "BOTTOMLEFT")
		f:SetPoint("TOPRIGHT", tab[key-1], "BOTTOMRIGHT")
	else
		f:SetPoint("TOPLEFT")
		f:SetPoint("TOPRIGHT")
	end
	f:SetHeight(30)
	f.highlight = f:CreateTexture()
	f.highlight:SetAllPoints()
	f.highlight:SetTexture(0.6,0.6,0.6)
	f.highlight:SetAlpha(0.3)
	f.highlight:Hide()
	f.selected = f:CreateTexture()
	f.selected:SetAllPoints()
	f.selected:SetTexture(0.8,0.8,0.8)
	f.selected:SetAlpha(0.6)
	f.selected:Hide()
	f:RegisterForClicks("LeftButtonUp")
	f:SetScript("OnEnter", function(self) if display.scrolls[1].selected ~= self then self.highlight:Show() end end)
	f:SetScript("OnLeave", function(self) self.highlight:Hide() end)
	f:SetScript("OnClick", function(self) for _, btn in ipairs(display.scrolls[2].entries) do btn:Hide() end for _, btn in ipairs(display.scrolls[3].entries) do btn:Hide() end self.clickFunc(self) if display.scrolls[1].selected then display.scrolls[1].selected.selected:Hide() end display.scrolls[1].selected = self self.selected:Show() end)
	f.label = f:CreateFontString()
	f.label:SetPoint("TOPLEFT", 5, -5)
	f.label:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
	f.label:SetTextColor(1,1,1)
	f.desc = f:CreateFontString()
	f.desc:SetPoint("BOTTOMLEFT", 5, 5)
	f.desc:SetFont("Fonts\\FRIZQT__.TTF", 8, "OUTLINE")
	f.desc:SetTextColor(1,1,1)
	display.scrolls[1].child:SetHeight(key*30)
	display.scrolls[1]:UpdateScrollChildRect()
	tab[key] = f
	return f
end})
display.scrolls[1]:SetScript("OnSizeChanged", function(self, width)
	self.child:SetWidth(width)
	self:UpdateScrollChildRect()
end)

local tempSortTable = {}
local sortName=function(a,b) return a.n < b.n end
do
	-- init primary categories
	local spellsBtn = display.scrolls[1].entries[1]
	spellsBtn.label:SetText("Spell list")
	spellsBtn.desc:SetText("View a list of spells found in logs")
	spellsBtn.clickFunc = function()
		for _, f in ipairs(display.scrolls[3].entries) do f:Hide() end
		table.wipe(tempSortTable)
		for spellID in pairs(CBMT.opt.indexedfights.spell) do
			tinsert(tempSortTable, {n=(GetSpellInfo(spellID)),s=spellID})
		end
		table.sort(tempSortTable,sortName)
		for i,data in ipairs(tempSortTable) do
			local btn = display.scrolls[3].entries[i]
			btn:SetOpen(data.s, 'spell')
			btn:Show()
		end
		display.scrolls[3].child:SetHeight(#tempSortTable*30)
		display.scrolls[3]:UpdateScrollChildRect()
		display.scrolls[2]:Hide()
		display.scrolls[1]:SetWidth((display:GetWidth()-30)/2)
		display.scrolls[3]:SetWidth((display:GetWidth()-30)/2)
		collectgarbage()
	end
	--[[local spellIndexBtn = display.scrolls[1].entries[2]
	spellIndexBtn.label:SetText("Indexed by spell")
	spellIndexBtn.desc:SetText("Search for a log by spell occurance")
	spellIndexBtn.clickFunc = function()
		for _, f in ipairs(display.scrolls[2].entries) do f:Hide() end
		table.wipe(tempSortTable)
		for spellID, listData in pairs(CBMT.opt.indexedfights.spell) do
			tinsert(tempSortTable,{s=spellID, n=(GetSpellInfo(spellID)), l=listData})
		end
		table.sort(tempSortTable,sortName)
		for i, data in ipairs(tempSortTable) do
			local btn = display.scrolls[2].entries[i]
			btn:SetList(data.n, data.l, data.s)
			btn:Show()
		end
		display.scrolls[2].child:SetHeight(#tempSortTable*20)
		display.scrolls[2]:UpdateScrollChildRect()
		display.scrolls[2]:Show()
		display.scrolls[1]:SetWidth((display:GetWidth()-40)/3)
		display.scrolls[3]:SetWidth((display:GetWidth()-40)/3)
		collectgarbage()
	end--]]
end

do
	local o = SlashCmdList["CBMT"]
	SlashCmdList["CBMT"] = function(input)
		if input == "spell" or input == "spelllogger" then
			display:Show()
			return
		end
		o(input)
	end
end

--[===[@debug@
SlashCmdList["NPCID"]=function(input) print(tonumber("0x"..(input:sub(7,10)))) end
SLASH_NPCID1 = "/npcid"

SlashCmdList["PURGEDB"] = function()
	table.wipe(CBMT.opt.indexedfights)
	print("Wiped fight indexes")
	for _, log in ipairs(CBMT.opt.logs) do
		local i=1
		local a = log[1]
		while a do
			if ignore_npcs[guid2npcid(a.srcGUID)] then
				print("Purged: "..a.src)
				tremove(log, i)
			else
				i=i+1
			end
			a=log[i]
		end
		log.names[UNKNOWN] = nil
		log.indexes = nil
		log.isIndexed = nil
	end
	print("Done!")
end
SLASH_PURGEDB1 = "/purgedb"
--@end-debug@]===]